# -*- coding: utf-8 -*-
"""base dumper

``dictdumper.dumper`` contains ``Dumper`` only, which is an
abstract base class for all dumpers, eg. HTML, JSON, PLIST,
Tree, and XML.

"""
import abc
import datetime
import functools
import os
import warnings
import sys


# Abstract Base Class of Dumpers
# Pre-define useful arguments and methods of dumpers


__all__ = ['Dumper']


ABCMeta = abc.ABCMeta
abstractmethod = abc.abstractmethod
abstractproperty = abc.abstractproperty


def deprecation(func):
    @functools.wraps(func)
    def wrapper(cls, *args, **kargs):
        if cls.__name__ in ('Dumper', 'JavaScript', 'XML'):
            warnings.warn(f'{cls.__name__} is deprecated', DeprecationWarning, stacklevel=2)
        return func(cls, *args, **kargs)
    return wrapper


class Dumper(object):
    """Abstract base class of all dumpers.

    Usage:
        >>> dumper = Dumper(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, file format of current dumper

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END

    """
    __metaclass__ = ABCMeta

    ##########################################################################
    # Attributes.
    ##########################################################################

    _sptr = os.SEEK_SET    # seek pointer
    _tctr = 1              # counter for tab level

    ##########################################################################
    # Properties.
    ##########################################################################

    # file format of current dumper
    @abstractproperty
    def kind(self):
        """File format of current dumper."""
        pass

    ##########################################################################
    # Data models.
    ##########################################################################

    # Not hashable
    __hash__ = None

    @deprecation
    def __new__(cls, fname, **kwargs):
        self = super().__new__(cls)
        self.object_hook = \
            kwargs.get('object_hook', cls.object_hook)
        return self

    def __init__(self, fname, **kwargs):
        if not os.path.isfile(fname):
            open(fname, 'w+').close()
        self._file = fname          # dump file name
        self._dump_header()         # initialise output file

    def __call__(self, value, *, name=None):
        with open(self._file, 'r+') as _file:
            self._append_value(value, _file, name)
            self._sptr = _file.tell()
            _file.write(self._hend)

    ##########################################################################
    # Utilities.
    ##########################################################################

    @classmethod
    def object_hook(cls, obj):
        """Check content type for function call."""
        if isinstance(obj, cls.__type__):
            return obj
        return repr(obj)

    def _dump_header(self):
        """Initially dump file heads and tails."""
        with open(self._file, 'w') as _file:
            _file.write(self._hsrt)
            self._sptr = _file.tell()
            _file.write(self._hend)

    @abstractmethod
    def _append_value(self, value, _file, _name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dunped
            * _file - FileIO, output file
            * _name - str, name of current content dict

        """
        pass



# -*- coding: utf-8 -*-
"""stream formatted output dumper

``dictdumper`` is an open source Python program works as a
stream formatted output dumper. Currently, it supports
following formats --

-  ``dictdumper.Dumper``
        abstract base class of all dumpers

-  ``dictdumper.JSON``
        dump JavaScript object notation (``JSON``)
        format file

-  ``dictdumper.PLIST``
        dump Apple property list (``PLIST``) format file

-  ``dictdumper.Tree``
        dump tree-view text (``TXT``) format file

-  ``dictdumper.XML``
        dump extensible Mmarkup language (``XML``) file;
        this is a deprecated base class

-  ``dictdumper.HTML``
        dump JavaScript file under ``Vue.js`` framework
        this class is deprecated due to grammar error

"""
# Base Class for jsFormat
from dictdumper.dumper import Dumper

# Utility Classes
from dictdumper.json import JSON
from dictdumper.plist import PLIST
from dictdumper.tree import Tree

# Deprecated / Base Classes
from dictdumper.html import JavaScript
from dictdumper.xml import XML


__all__ = ['JSON', 'PLIST', 'Tree']



# -*- coding: utf-8 -*-
"""dumper a JavaScript file (DEPRECARED)

    Note that this file is deprecated.

``dictdumper.html`` contains ``HTML`` only, which dumpers a
JavaScript file according to ``Vue.js`` framework. However,
due to errors in grammar, the output file won't work, thus
it is now deprecated. Usage sample is described as below.

    >>> dumper = JavaScript(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
import collections
import os
import textwrap


# Writer for JavaScript files
# Dump a JavaScript file for PCAP analyser


from dictdumper.json import JSON


__all__ = ['JavaScript']


# head
_HEADER_START = '''\
// demo data
var data = {
'''


# tail
_HEADER_END = """
}

// define the item component
Vue.component('item', {
  template: '#item-template',
  props: {
    model: Object
  },
  data: function () {
    return {
      open: false
    }
  },
  computed: {
    isFolder: function () {
      return this.model.children &&
        this.model.children.length
    }
  },
  methods: {
    toggle: function () {
      if (this.isFolder) {
        this.open = !this.open
      }
    },
    changeType: function () {
      if (!this.isFolder) {
        Vue.set(this.model, 'children', [])
        this.addChild()
        this.open = true
      }
    },
    addChild: function () {
      this.model.children.push({
        name: 'new stuff'
      })
    }
  }
})

// boot up the demo
var demo = new Vue({
  el: '#demo',
  data: {
    treeData: data
  }
})
"""


class JavaScript(JSON):
    """Dump JavaScript file under `Vue.js` framework.

    Usage:
        >>> dumper = JavaScript(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'js'

    Methods:
        * object_hook - default/customised object hooks

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END
        * _vctr - dict, value counter dict

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'js'

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END
    _vctr = collections.defaultdict(int)    # value counter dict



# -*- coding: utf-8 -*-
"""dumper a JSON file

``dictdumper.json`` contains ``JSON`` only, which dumpers a
JavaScript object notation (JSON) file. Usage sample is
described as below.

    >>> dumper = JSON(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
import collections
import datetime
import os
import textwrap


# Writer for JSON files
# Dump a JSON file for PCAP analyser


from dictdumper.dumper import Dumper


# head
_HEADER_START = '{\n'


# tail
_HEADER_END = '\n}'


# magic types
_MAGIC_TYPES = collections.defaultdict(
    lambda : (lambda self, text, file: self._append_string(text, file)), dict(
    # string
    str = lambda self, text, file: self._append_string(text, file),     

    # bytes
    bytes = lambda self, text, file: self._append_bytes(text, file),    
    bytearray = lambda self, text, file: self._append_bytes(text, file),
    memoryview = lambda self, text, file: self._append_bytes(text, file),

    # date
    datetime = lambda self, text, file: self._append_date(text, file),  

    # number
    int = lambda self, text, file: self._append_number(text, file),     
    float = lambda self, text, file: self._append_number(text, file),   

    # object
    dict = lambda self, text, file: self._append_object(text, file),     

    # array
    list = lambda self, text, file: self._append_array(text, file),     
    tuple = lambda self, text, file: self._append_array(text, file),    
    range = lambda self, text, file: self._append_array(text, file),
    set = lambda self, text, file: self._append_array(text, file),
    frozenset = lambda self, text, file: self._append_array(text, file),

    # bool
    bool = lambda self, text, file: self._append_bool(text, file),      

    # null
    NoneType = lambda self, text, file: self._append_null(text, file),  
))


class JSON(Dumper):
    """Dump JavaScript object notation (JSON) format file.

    Usage:
        >>> dumper = JSON(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'json'

    Methods:
        * object_hook - default/customised object hooks

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END
        * _vctr - dict, value counter dict

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    Terminology:
        object    ::=  "{}" | ("{" members "}")
        members   ::=  pair | (pair "," members)
        pair      ::=  string ":" value
        array     ::=  "[]" | ("[" elements "]")
        elements  ::=  value | (value "," elements)
        value     ::=  string | number | object
                         | array | true | false | null

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'json'

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = (
        str,                                    # stinrg
        bool,                                   # bool
        dict,                                   # object
        datetime.date,                          # date
        int, float, complex,                    # number
        type(None),                             # null
        bytes, bytearray, memoryview,           # bytes
        list, tuple, range, set, frozenset,     # array
    )

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END
    _vctr = collections.defaultdict(int)    # value counter dict

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _append_value(self, value, _file, _name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file
            * _name - str, name of current content dict

        """
        _tabs = '\t' * self._tctr
        _cmma = ',\n' if self._vctr[self._tctr] else ''
        _keys = '{cmma}{tabs}"{name}" :'.format(cmma=_cmma, tabs=_tabs, name=_name)

        _file.seek(self._sptr, os.SEEK_SET)
        _file.write(_keys)

        self._vctr[self._tctr] += 1
        self._append_object(value, _file)

    ##########################################################################
    # Functions.
    ##########################################################################

    def _append_array(self, value, _file):
        """Call this function to write array contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _labs = ' ['
        _file.write(_labs)

        self._tctr += 1

        for _item in value:
            _cmma = ',' if self._vctr[self._tctr] else ''
            _file.write(_cmma)

            self._vctr[self._tctr] += 1

            _item = self.object_hook(_item)
            _type = type(_item).__name__
            _MAGIC_TYPES[_type](self, _item, _file)

        self._vctr[self._tctr] = 0
        self._tctr -= 1

        _labs = ' ]'
        _file.write(_labs)

    def _append_object(self, value, _file):
        """Call this function to write object contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _labs = ' {'
        _file.write(_labs)
        self._tctr += 1

        for (_item, _text) in value.items():
            _tabs = '\t' * self._tctr
            _cmma = ',' if self._vctr[self._tctr] else ''
            _keys = '{cmma}\n{tabs}"{item}" :'.format(cmma=_cmma, tabs=_tabs, item=_item)
            _file.write(_keys)

            self._vctr[self._tctr] += 1

            _text = self.object_hook(_text)
            _type = type(_text).__name__
            _MAGIC_TYPES[_type](self, _text, _file)

        self._vctr[self._tctr] = 0
        self._tctr -= 1
        _tabs = '\t' * self._tctr
        _labs = '\n{tabs}{}'.format('}', tabs=_tabs)
        _file.write(_labs)

    def _append_string(self, value, _file):
        """Call this function to write string contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = str(value).replace('"', '\\"')
        _labs = ' "{text}"'.format(text=_text)
        _file.write(_labs)

    def _append_bytes(self, value, _file):
        """Call this function to write bytes contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        # binascii.b2a_base64(value) -> plistlib.Data
        # binascii.a2b_base64(Data) -> value(bytes)

        _text = ' '.join(textwrap.wrap(value.hex(), 2))
        # _data = [H for H in iter(
        #         functools.partial(io.StringIO(value.hex()).read, 2), '')
        #         ]  # to split bytes string into length-2 hex string list
        _labs = ' "{text}"'.format(text=_text)
        _file.write(_labs)

    def _append_date(self, value, _file):
        """Call this function to write date contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = value.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        _labs = ' "{text}"'.format(text=_text)
        _file.write(_labs)

    def _append_number(self, value, _file):
        """Call this function to write number contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = value
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_bool(self, value, _file):
        """Call this function to write bool contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = 'true' if value else 'false'
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_null(self, value, _file):
        """Call this function to write null contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = 'null'
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)



# -*- coding: utf-8 -*-
"""dumper a XML file (DEPRECATED)

    Note that this file is deprecated.

``dictdumper.xml`` contains ``XML`` only, which dumpers an
extensible markup language (XML) format file. However, due
to lack of supported ``dtd``, the output file is currently
meaningless, thus it is now deprecated. Usage sample is
described as below.

    >>> dumper = XML(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
# TODO: Supports more `dtd`s of XML.


import os
import textwrap


# Dumper for XML files
# Write a XML file for PCAP analyser


from dictdumper.dumper import Dumper


# head
_HEADER_START = '''\
<?xml version="1.0" encoding="UTF-8"?>
<content>
'''


# tail
_HEADER_END = '''\
</content>
'''


class XML(Dumper):
    """Dump extensible markup language (XML) file.

    Note:
        This is a base dumper for XML format. No `dtd` supported.

    Usage:
        >>> dumper = XML(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'plist'

    Methods:
        * object_hook - default/customised object hooks

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END

    Methods:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'xml'

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END



#!/usr/bin/python3
# -*- coding: utf-8 -*-


import setuptools


# README
with open('./README.md', 'r') as file:
    long_desc = file.read()


# version string
__version__ = '0.6.0'


# set-up script for pip distribution
setuptools.setup(
    name = 'dictdumper',
    version = __version__,
    author = 'Jarry Shaw',
    author_email = 'jarryshaw@icloud.com',
    url = 'https://github.com/JarryShaw/dictdumper#dictdumper',
    license = 'GNU General Public License v3 (GPLv3)',
    keywords = 'formatting dumper stream',
    description = 'Python dict formatted dumper.',
    long_description = long_desc,
    long_description_content_type='text/markdown',
    install_requires = ['setuptools'],
    python_requires = '>=3.0',
    py_modules = ['dictdumper'],
    packages = ['dictdumper'],
    package_data = {
        '': [
            'LICENSE',
            'README.md',
        ],
    },
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Networking',
        'Topic :: Utilities',
    ]
)



# -*- coding: utf-8 -*-
"""dumper a tree-view file

``dictdumper.tree`` contains ``Tree`` only, which dumpers a
tree-view text (TXT) format file. Usage sample is described
as below.

    >>> dumper = Tree(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
import collections
import datetime
import os
import textwrap


# Writer for treeview text files
# Dump a TEXT file for PCAP analyser


from dictdumper.dumper import Dumper


# headers
_HEADER_START = 'PCAP File Tree-View Format\n'   # head
_HEADER_END = ''                               # tail


# templates
_TEMP_BRANCH = '  |   '  # branch
_TEMP_SPACES = '      '  # space


# magic types
_MAGIC_TYPES = collections.defaultdict(
    lambda : (lambda self, text, file: self._append_string(text, file)), dict(
    # branch
    dict = lambda self, text, file: self._append_branch(text, file),            

    # array
    set = lambda self, text, file: self._append_array(text, file),
    list = lambda self, text, file: self._append_array(text, file),             
    tuple = lambda self, text, file: self._append_array(text, file),            
    range = lambda self, text, file: self._append_array(text, file),            
    frozenset = lambda self, text, file: self._append_array(text, file),

    # string
    str = lambda self, text, file: self._append_string(text, file),             

    # date
    datetime = lambda self, text, file: self._append_date(text, file),          

    # bytes
    bytes = lambda self, text, file: self._append_bytes(text, file),            
    bytearray = lambda self, text, file: self._append_bytes(text, file),        
    memoryview = lambda self, text, file: self._append_bytes(text, file),       

    # number
    int = lambda self, text, file: self._append_number(text, file),             
    float = lambda self, text, file: self._append_number(text, file),           
    complex = lambda self, text, file: self._append_number(text, file),         

    # True | False
    bool = lambda self, text, file: self._append_bool(text, file),              

    # N/A
    NoneType = lambda self, text, file: self._append_none(text, file),          
))


class Tree(Dumper):
    """Dump a tree-view text (TXT) format file.

    Usage:
        >>> dumper = Tree(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'plist'

    Methods:
        * object_hook - default/customised object hooks

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END
        * _bctr - dict, blank branch counter dict

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    Terminology:
        value   ::=  branch | array | string | number | bool | N/A
        string
          |-- string
          |     |-- string -> value
          |     |-- string
          |     |     |-- string -> value
          |     |     |-- string -> value
          |     |-- string -> value
          |     |-- string -> value
          |           |-- string -> value
          |           |-- string -> value
          |-- string -> value, value, value
          |-- string -> True
          |-- string -> False
          |-- string -> N/A
          |-- string -> value
          |-- string -> value

    """
    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = (
        str,                                    # string
        bool,                                   # bool
        dict,                                   # branch
        type(None),                             # none
        datetime.date,                          # date
        int, float, complex,                    # number
        bytes, bytearray, memoryview,           # bytes
        list, tuple, range, set, frozenset,     # array
    )

    ##########################################################################
    # Attributes.
    ##########################################################################

    _tctr = -1
    _hsrt = _HEADER_START
    _hend = _HEADER_END

    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'txt'

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _append_value(self, value, _file, _name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file
            * _name - str, name of current content dict

        """
        _keys = '\n' + _name + '\n'
        _file.seek(self._sptr, os.SEEK_SET)
        _file.write(_keys)

        self._bctr = collections.defaultdict(int)    # blank branch counter dict
        self._append_branch(value, _file)

    def _append_array(self, value, _file):
        """Call this function to write array contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        if not value:
            return self._append_none(None, _file)

        _bptr = ''
        _tabs = ''
        _tlen = len(value) - 1
        if _tlen:
            _bptr = '  |-->'
            for _ in range(self._tctr + 1):
                _tabs += _TEMP_SPACES if self._bctr[_] else _TEMP_BRANCH

        for (_nctr, _item) in enumerate(value):
            _text = '{tabs}{bptr}'.format(tabs=_tabs, bptr=_bptr)
            _file.write(_text)

            _item = self.object_hook(_item)
            _type = type(_item).__name__
            _MAGIC_TYPES[_type](self, _item, _file)

            _suff = '\n' if _nctr < _tlen else ''
            _file.write(_suff)

    def _append_branch(self, value, _file):
        """Call this function to write branch contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        if not value:
            return self._append_none(None, _file)

        self._tctr += 1
        _vlen = len(value)
        for (_vctr, (_item, _text)) in enumerate(value.items()):
            _text = self.object_hook(_text)
            _type = type(_text).__name__

            flag_dict = (_type == 'dict')
            flag_tuple = (_type == 'tuple' and len(_text) > 1)
            flag_bytes = (_type == 'bytes' and len(_text) > 16)
            if any((flag_dict, flag_tuple, flag_bytes)):
                _pref = '\n'
            else:
                _pref = ' ->'

            _labs = ''
            for _ in range(self._tctr):
                _labs += _TEMP_SPACES if self._bctr[_] else _TEMP_BRANCH

            _keys = '{labs}  |-- {item}{pref}'.format(labs=_labs, item=_item, pref=_pref)
            _file.write(_keys)

            if _vctr == _vlen - 1:
                self._bctr[self._tctr] = 1

            _MAGIC_TYPES[_type](self, _text, _file)

            _suff = '' if _type == 'dict' else '\n'
            _file.write(_suff)

        self._bctr[self._tctr] = 0
        self._tctr -= 1

    def _append_string(self, value, _file):
        """Call this function to write string contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        if not value:
            return self._append_none(None, _file)

        _text = value
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_bytes(self, value, _file):
        """Call this function to write bytes contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        # binascii.b2a_base64(value) -> plistlib.Data
        # binascii.a2b_base64(Data) -> value(bytes)
        if not value:
            return self._append_none(None, _file)

        if len(value) > 16:
            _tabs = ''
            for _ in range(self._tctr + 1):
                _tabs += _TEMP_SPACES if self._bctr[_] else _TEMP_BRANCH

            _list = []
            for (_ictr, _item) in enumerate(textwrap.wrap(value.hex(), 32)):
                _bptr = '       ' if _ictr else '  |--> '
                _text = ' '.join(textwrap.wrap(_item, 2))
                _item = '{tabs}{bptr}{text}'.format(tabs=_tabs, bptr=_bptr, text=_text)
                _list.append(_item)
            _labs = '\n'.join(_list)
        else:
            _text = ' '.join(textwrap.wrap(value.hex(), 2))
            _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_date(self, value, _file):
        """Call this function to write date contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = value.strftime('%Y-%m-%d %H:%M:%S.%f')
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_number(self, value, _file):
        """Call this function to write number contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = value
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_bool(self, value, _file):
        """Call this function to write bool contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = 'True' if value else 'False'
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)

    def _append_none(self, value, _file):
        """Call this function to write none contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _text = 'NIL'
        _labs = ' {text}'.format(text=_text)
        _file.write(_labs)



# -*- coding: utf-8 -*-
"""dumper a PLIST file

``dictdumper.plist`` contains ``PLIST`` only, which dumpers an
Apple property list (PLIST) file. Usage sample is described
as below.

    >>> dumper = PLIST(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
import collections
import datetime
import os
import textwrap


# Dumper for PLIST files
# Write a macOS Property List file


from dictdumper.xml import XML


# head
_HEADER_START = '''\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
'''


# tail
_HEADER_END = '''\
</dict>
</plist>
'''


# magic types
_MAGIC_TYPES = collections.defaultdict(
    lambda : (lambda self, text, file: self._append_string(text, file)), dict(
    # array
    list = lambda self, text, file: self._append_array(text, file),         
    tuple = lambda self, text, file: self._append_array(text, file),        
    range = lambda self, text, file: self._append_array(text, file),
    set = lambda self, text, file: self._append_array(text, file),
    frozenset = lambda self, text, file: self._append_array(text, file),

    # dict
    dict = lambda self, text, file: self._append_dict(text, file),          

    # string
    str = lambda self, text, file: self._append_string(text, file),         

    # data
    bytes = lambda self, text, file: self._append_data(text, file),         
    bytearray = lambda self, text, file: self._append_data(text, file),
    memoryview = lambda self, text, file: self._append_data(text, file),

    # date
    datetime = lambda self, text, file: self._append_date(text, file),      

    # integer
    int = lambda self, text, file: self._append_integer(text, file),        

    # real
    float = lambda self, text, file: self._append_real(text, file),         

    # true | false
    bool = lambda self, text, file: self._append_bool(text, file),          
))


class PLIST(XML):
    """Dump Apple property list (PLIST) format file.

    Usage:
        >>> dumper = PLIST(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'plist'

    Methods:
        * object_hook - default/customised object hooks

    Attributes:
        * _file - FileIO, output file
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _append_value - call this function to write contents

    Terminology:
        value    ::=  array | dict | string | data
                        | date | integer | real | bool
        array    ::=  "<array>" value* "</array>"
        dict     ::=  "<dict>" ("<key>" str "</key>" value)* "</dict>"
        string   ::=  "<string>" str "</string>"
        data     ::=  "<data>" bytes "</data>"
        date     ::=  "<date>" datetime "</date>"
        integer  ::=  "<integer>" int "</integer>"
        real     ::=  "<real>" float "</real>"
        bool     ::=  "<true/>" | "<false/>"

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'plist'

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = (
        str,                                    # stinrg
        bool,                                   # bool
        dict,                                   # dict
        datetime.date,                          # date
        int,                                    # integer
        float,                                  # real
        bytes, bytearray, memoryview,           # data
        list, tuple, range, set, frozenset,     # array
    )

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _append_value(self, value, _file, _name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file
            * _name - str, name of current content dict

        """
        _tabs = '\t' * self._tctr
        _keys = '{tabs}<key>{name}</key>\n'.format(tabs=_tabs, name=_name)
        _file.seek(self._sptr, os.SEEK_SET)
        _file.write(_keys)

        self._append_dict(value, _file)

    def _append_array(self, value, _file):
        """Call this function to write array contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _labs = '{tabs}<array>\n'.format(tabs=_tabs)
        _file.write(_labs)
        self._tctr += 1

        for _item in value:
            if _item is None:   continue
            _item = self.object_hook(_item)
            _type = type(_item).__name__
            _MAGIC_TYPES[_type](self, _item, _file)

        self._tctr -= 1
        _tabs = '\t' * self._tctr
        _labs = '{tabs}</array>\n'.format(tabs=_tabs)
        _file.write(_labs)

    def _append_dict(self, value, _file):
        """Call this function to write dict contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _labs = '{tabs}<dict>\n'.format(tabs=_tabs)
        _file.write(_labs)
        self._tctr += 1

        for (_item, _text) in value.items():
            if _text is None:   continue

            _tabs = '\t' * self._tctr
            _keys = '{tabs}<key>{item}</key>\n'.format(tabs=_tabs, item=_item)
            _file.write(_keys)

            _text = self.object_hook(_text)
            _type = type(_text).__name__
            _MAGIC_TYPES[_type](self, _text, _file)

        self._tctr -= 1
        _tabs = '\t' * self._tctr
        _labs = '{tabs}</dict>\n'.format(tabs=_tabs)
        _file.write(_labs)

    def _append_string(self, value, _file):
        """Call this function to write string contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _text = value
        _labs = '{tabs}<string>{text}</string>\n'.format(tabs=_tabs, text=_text)
        _file.write(_labs)

    def _append_data(self, value, _file):
        """Call this function to write data contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        # binascii.b2a_base64(value) -> plistlib.Data
        # binascii.a2b_base64(Data) -> value(bytes)

        _tabs = '\t' * self._tctr
        _text = value.hex() # str(value)[2:-1]
        _labs = '{tabs}<data>{text}</data>\n'.format(tabs=_tabs, text=_text)
        # _labs = '{tabs}<data>\n'.format(tabs=_tabs)

        # _list = []
        # for _item in textwrap.wrap(value.hex(), 32):
        #     _text = ' '.join(textwrap.wrap(_item, 2))
        #     _item = '{tabs}\t{text}'.format(tabs=_tabs, text=_text)
        #     _list.append(_item)
        # _labs += '\n'.join(_list)

        # _data = [H for H in iter(
        #         functools.partial(io.StringIO(value.hex()).read, 2), '')
        #         ]  # to split bytes string into length-2 hex string list
        # _labs += '\n{tabs}</data>\n'.format(tabs=_tabs)
        _file.write(_labs)

    def _append_date(self, value, _file):
        """Call this function to write date contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _text = value.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        _labs = '{tabs}<date>{text}</date>\n'.format(tabs=_tabs, text=_text)
        _file.write(_labs)

    def _append_integer(self, value, _file):
        """Call this function to write integer contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _text = value
        _labs = '{tabs}<integer>{text}</integer>\n'.format(tabs=_tabs, text=_text)
        _file.write(_labs)

    def _append_real(self, value, _file):
        """Call this function to write real contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _text = value
        _labs = '{tabs}<real>{text}</real>\n'.format(tabs=_tabs, text=_text)
        _file.write(_labs)

    def _append_bool(self, value, _file):
        """Call this function to write bool contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file

        """
        _tabs = '\t' * self._tctr
        _text = '<true/>' if value else '<false/>'
        _labs = '{tabs}{text}\n'.format(tabs=_tabs, text=_text)
        _file.write(_labs)
